from asyncio import subprocess
from pathlib import Path
import click
from pydantic import BaseModel, Field
import yaml
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend
import typer
from arkitekt_server.config import (
    ArkitektServerConfig,
    BaseServiceConfig,
    RekuestConfig,
    KeyPair,
    User,
)
from typer.core import TyperGroup
from arkitekt_server.diff import run_dry_run_diff

ASCI_LOGO = r"""
 █████╗ ██████╗ ██╗  ██╗██╗████████╗███████╗██╗  ██╗████████╗
██╔══██╗██╔══██╗██║ ██╔╝██║╚══██╔══╝██╔════╝██║ ██╔╝╚══██╔══╝
███████║██████╔╝█████╔╝ ██║   ██║   █████╗  █████╔╝    ██║   
██╔══██║██╔══██╗██╔═██╗ ██║   ██║   ██╔══╝  ██╔═██╗    ██║   
██║  ██║██║  ██║██║  ██╗██║   ██║   ███████╗██║  ██╗   ██║   
╚═╝  ╚═╝╚═╝  ╚═╝╚═╝  ╚═╝╚═╝   ╚═╝   ╚══════╝╚═╝  ╚═╝   ╚═╝   
"""


app = typer.Typer()


auth_app = typer.Typer()
app.add_typer(
    auth_app, name="auth", help="Authentication related settings and commands"
)


init_app = typer.Typer()
app.add_typer(
    init_app, name="init", help="Initialize an Arkitekt deployment configuration"
)


build_app = typer.Typer()
app.add_typer(build_app, name="build", help="Build deployments for Arkitekt server")


user_app = typer.Typer()
auth_app.add_typer(user_app, name="user", help="Standard user management commands")


group_app = typer.Typer()
auth_app.add_typer(user_app, name="group", help="Group management commands")


admin_app = typer.Typer()
auth_app.add_typer(admin_app, name="admin", help="Admin users management commands")


service_app = typer.Typer()
app.add_typer(service_app, name="service", help="Service management commands")


class YamlFile(BaseModel):
    version: str
    config: ArkitektServerConfig


def update_or_create_yaml_file(file_path: str, new_config: ArkitektServerConfig):
    """Update or create a YAML file with the given data."""

    with open("arkitekt_server_config.yaml", "w") as f:
        yaml_data = YamlFile(version="1.0", config=new_config)
        yaml.dump(yaml_data.model_dump(), f, default_flow_style=False)

    # create gitignore file if it does not exist
    try:
        with open(".gitignore", "r") as f:
            gitignore_content = f.read()
    except FileNotFoundError:
        gitignore_content = ""

    if "arkitekt_server_config.yaml" not in gitignore_content:
        with open(".gitignore", "a") as f:
            f.write("\narkitekt_server_config.yaml\n")
            f.write("# Arkitekt server config file\n")
            f.write(
                "# This file is automatically generated and should not be modified manually.\n"
            )
            f.write(
                "# It contains sensitive information and should not be committed to version control.\n"
            )


def load_or_create_yaml_file(file_path: str) -> ArkitektServerConfig:
    """Load or create a YAML file with default configuration."""
    try:
        with open("arkitekt_server_config.yaml", "r") as f:
            data = yaml.safe_load(f)
            return ArkitektServerConfig(**data["config"])
    except FileNotFoundError:
        # If the file does not exist, create a new one with default values
        default_config = ArkitektServerConfig()
        update_or_create_yaml_file(file_path, default_config)
        return default_config


def load_yaml_file(file_path: str) -> ArkitektServerConfig:
    """Load a YAML file and return the configuration."""
    try:
        with open(file_path, "r") as f:
            data = yaml.safe_load(f)
            return ArkitektServerConfig(**data["config"])
    except FileNotFoundError:
        raise FileNotFoundError(f"Configuration file '{file_path}' not found.")
    except yaml.YAMLError as e:
        raise ValueError(f"Error parsing YAML file: {e}")


@init_app.command()
def default():
    """Build commands for Arkitekt server.

    Creates a config that can be used to run the Arkitekt server.



    """

    # Create a default configuration file if it doesn't exist
    config = ArkitektServerConfig()

    print("Creating default configuration file for Arkitekt server...")
    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    # load the yaml file


@init_app.command()
def dev():
    """Build commands for Arkitekt server.


    Creates a config that can be used to run the Arkitekt server.



    """

    # Create a default configuration file if it doesn't exist
    config = ArkitektServerConfig()

    config.rekuest.mount_github = True
    config.kabinet.mount_github = True
    config.mikro.mount_github = True
    config.fluss.mount_github = True
    config.deployer.mount_github = True
    config.kraph.mount_github = True
    print("Creating default configuration file for Arkitekt server...")
    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    # load the yaml file


@init_app.command()
def minimal():
    """Build a minimal configuration for the Arkitekt server.

    This command creates a minimal configuration file that can be used to run the Arkitekt server.
    It includes only the most essential settings and services required to start the server.

    """

    # Create a default configuration file if it doesn't exist
    config = ArkitektServerConfig()

    print("Creating default configuration file for Arkitekt server...")
    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    # load the yaml file


@user_app.command()
def add():
    """Add a new user to the Arkitekt server configuration."""

    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    username = click.prompt("Enter the username for the user", type=str)

    password = click.prompt(
        "Enter the password for the user", type=str, hide_input=True
    )

    email = click.prompt("Enter the email for the user", type=str, default=None)

    config.users.append(
        User(
            username=username,
            password=password,
            email=email,
        )
    )
    # Here you would typically hash the password and store it securely
    # For simplicity, we will just store it in plain text (not recommended for production)

    update_or_create_yaml_file("arkitekt_server_config.yaml", config)


@user_app.command()
def remove(name: str):
    """Remove a user from the Arkitekt server configuration."""
    print("Removing user is not implemented yet.")


@service_app.command()
def rekuest(enable: bool = True):
    """Add the Rekuest service to the Arkitekt server configuration."""

    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    rekuest_service = RekuestConfig()

    config.kabinet.enabled = enable

    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    click.echo("Rekuest service added with default configuration.")


@service_app.command()
def mikro(
    enable: bool = True,
):
    """Add the Mikro service to the Arkitekt server configuration."""

    config = load_or_create_yaml_file("arkitekt_server_config.yaml")
    config.kabinet.enabled = enable

    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    click.echo("Mikro service added with default configuration.")


@service_app.command()
def kabinet(enable: bool = True):
    """Add the Kabinet service to the Arkitekt server configuration."""

    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    config.kabinet.enabled = enable

    update_or_create_yaml_file("arkitekt_server_config.yaml", config)

    click.echo("Kabinet service added with default configuration.")


@build_app.command()
def docker(path: Path = Path(".")):
    """Build the Docker image for the Arkitekt server."""

    # load the yaml file
    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    run_dry_run_diff(config, path, allow_deletes=False)


@build_app.command()
def kubernetes(path: Path = Path(".")):
    """Build the Docker image for the Arkitekt server."""

    # load the yaml file
    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    raise Exception("Kubernetes support is not implemented yet.")


@app.command()
def migrate():
    """Migrate the Arkitekt server configuration."""

    # load the yaml file
    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    # Here you would typically perform migration tasks, such as updating the schema or data format.
    # For simplicity, we will just print the current configuration.


@app.command()
def update():
    """Update the Arkitekt server configuration."""

    # load the yaml file
    config = load_or_create_yaml_file("arkitekt_server_config.yaml")

    click.echo(f"Updating Arkitekt server with configuration: {config.json(indent=2)}")

    # Here you would typically perform update tasks, such as applying patches or changes to the configuration.
    # For simplicity, we will just print the current configuration.


@app.command()
def up():
    """Start the Arkitekt server."""

    # load the yaml file
    config = load_yaml_file("arkitekt_server_config.yaml")

    subprocess.run(["docker", "compose", "up", "-d"], check=True)


def main():
    """Main entry point for the Arkitekt server CLI."""
    app()
